# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import


class Span(object):
    """
    Span represents a unit of work executed on behalf of a trace. Examples of
    spans include a remote procedure call, or a in-process method call to a
    sub-component. A trace is required to have a single, top level "root"
    span, and zero or more children spans, which in turns can have their own
    children spans, thus forming a tree structure.

    Span implements a Context Manager API that allows the following usage:

    .. code-block:: python

        span = tracer.start_trace(operation_name='go_fishing')
        with span:
            call_some_service()

    In this case it's not necessary to call span.finish()
    """

    def __init__(self, trace_context):
        self.trace_context = trace_context

    def __enter__(self):
        """Invoked when span is used as a context manager.

        :return: returns the Span itself
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Ends context manager and calls finish() on the span.

        If exception has occurred during execution, it is automatically added
        as a tag to the span.
        """
        if exc_type:
            self.error('python.exception', {'type': exc_type,
                                            'val': exc_val,
                                            'tb': exc_tb})
        self.finish()

    def start_child(self, operation_name, tags=None):
        """Denotes the beginning of a subordinate unit of work.

        As this is a no-op implementation, it actually returns itself as the
        child span. Real implementations should create a new span.

        :param operation_name: name of the operation represened by the new
            span from the perspective of the current service.
        :param tags: optional dictionary of Span Tags. The caller is
            expected to give up ownership of that dictionary, because the
            Tracer may use it as is to avoid extra data copying.

        :return: Returns a new child Span in "started" state.
        """
        return self

    def finish(self):
        """Indicates that the work represented by this span has been completed
        or terminated, and is ready to be sent to the Reporter.

        If any tags / logs need to be added to the span, it should be done
        before calling finish(), otherwise they may be ignored.
        """
        pass

    def add_tag(self, key, value):
        """Attaches a key/value pair to the span. More than one value can be
        added for the same key.

        The set of supported value types is implementation specific. It is the
        responsibility of the actual tracing system to know how to serialize
        and record the values.

        :param key: key or name of the tag. Must be a string.
        :param value: value of the tag.

        :return: Returns the Span itself, for call chaining.
        :rtype: Span
        """
        return self

    def info(self, message, *args):
        """Logs an event/message against the span, with the current timestamp.

        :param message: a format string that can refer to fields
            in the args payload.
        :param args: arbitrary payload
        :return: returns the span itself, for chaining the calls
        """
        return self

    def error(self, message, *args):
        """Same as info(), but for errors.

        :param message: a format string that can refer to fields
            in the args payload.
        :param args: arbitrary payload
        :return: returns the span itself, for chaining the calls
        """
        return self



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import sys
from concurrent.futures import Future
from tornado import gen, ioloop


def get_io_loop(channel):
    """
    Extracts IOLoop from TChannel instance, or tries to get the current one,
    but without starting a new one.
    :param channel:
    :return:
    """
    if hasattr(channel, '_threadloop') and \
            hasattr(channel._threadloop, '_io_loop'):
        return channel._threadloop._io_loop
    return ioloop.IOLoop.current(instance=False)


def submit(fn, io_loop, *args, **kwargs):
    """Submit Tornado Coroutine to IOLoop.current().

    :param fn: Tornado Coroutine to execute
    :param io_loop: Tornado IOLoop where to schedule the coroutine
    :param args: Args to pass to coroutine
    :param kwargs: Kwargs to pass to coroutine
    :returns concurrent.futures.Future: future result of coroutine
    """
    future = Future()

    def execute():
        """Executes fn on the IOLoop."""
        try:
            result = gen.maybe_future(fn(*args, **kwargs))
        except Exception:
            # The function we ran didn't return a future and instead raised
            # an exception. Let's pretend that it returned this dummy
            # future with our stack trace.
            f = gen.Future()
            f.set_exc_info(sys.exc_info())
            on_done(f)
        else:
            result.add_done_callback(on_done)

    def on_done(tornado_future):
        """Sets tornado.Future results to the concurrent.Future."""

        exception = tornado_future.exception()
        if not exception:
            future.set_result(tornado_future.result())
        else:
            future.set_exception(exception)

    io_loop.add_callback(execute)

    return future


def future_result(result):
    future = Future()
    future.set_result(result)
    return future


def future_exception(exception):
    future = Future()
    future.set_exception(exception)
    return future



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
import opentracing
from .context import TraceContext


class TraceContextEncoder(opentracing.TraceContextEncoder):
    """
    Implements opentracing.TraceContextEncoder.
    """
    def __init__(self, trace_id_header, trace_attributes_header_prefix):
        self.prefix = trace_attributes_header_prefix.lower()
        self.prefix_length = len(trace_attributes_header_prefix)
        self.trace_id_header = trace_id_header.lower().replace('_', '-')

    def trace_context_to_binary(self, trace_context):
        raise NotImplementedError('trace_context_to_binary')

    def trace_context_to_text(self, trace_context):
        """Converts trace context to a pair of dict's representing
        separately span identity and Trace Attributes.

        :param trace_context: trace context to encode
        :type trace_context: TraceContext

        :return: a pair of str dicts, first representing the span identity,
            the second representing Trace Attributes.
        """
        ctx_id = TraceContextEncoder.id_to_string(trace_context)
        span_id_out = {self.trace_id_header: ctx_id}
        trace_attributes = trace_context.trace_attributes_as_dict()
        trace_attributes_out = None
        if trace_attributes is not None:
            trace_attributes_out = {}
            for key, value in trace_attributes.iteritems():
                trace_attributes_out['%s%s' % (self.prefix, key)] = value
        return span_id_out, trace_attributes_out

    @staticmethod
    def id_to_string(trace_context):
        """
        Serialize span ID from the trace context to a string
        ``{trace_id}:{span_id}:{parent_id}:{flags}.``
        Number are encoded as variable-length lower-case hex strings.
        If parent_id is None, it is written as 0.
        """
        parent_id = trace_context.parent_id \
            if trace_context.parent_id is not None \
            else 0L
        return "{:x}:{:x}:{:x}:{:x}".format(trace_context.trace_id,
                                            trace_context.span_id,
                                            parent_id,
                                            trace_context.flags)


class TraceContextDecoder(opentracing.TraceContextDecoder):

    def __init__(self, trace_id_header, trace_attributes_header_prefix):
        self.prefix = trace_attributes_header_prefix.lower()
        self.prefix_length = len(trace_attributes_header_prefix)
        self.trace_id_header = trace_id_header.lower().replace('_', '-')

    def trace_context_from_binary(self, trace_context_id,
                                  trace_attributes):
        raise NotImplementedError('trace_context_from_binary')

    def trace_context_from_text(self, trace_context_id,
                                trace_attributes):
        ctx_id = trace_context_id.get(self.trace_id_header, None)
        if ctx_id is None:
            return None
        trace_context = TraceContextDecoder.\
            trace_context_from_string(value=ctx_id)
        if trace_context is None:
            return None
        trace_attributes_out = None
        if trace_attributes:
            for key, value in trace_attributes.iteritems():
                uc_key = key.lower()
                if uc_key.startswith(self.prefix):
                    attr_key = key[self.prefix_length:]
                    if trace_attributes_out is None:
                        trace_attributes_out = {attr_key.lower(): value}
                    else:
                        trace_attributes_out[attr_key.lower()] = value
        if trace_attributes_out is not None:
            trace_context.append_trace_attributes(trace_attributes_out)
        return trace_context

    @staticmethod
    def trace_context_from_string(value):
        """
        Decodes span ID from a string into a TraceContext.
        Returns None if the string value is malformed.
        """
        if type(value) is list and len(value) > 0:
            # sometimes headers are presented as arrays of values
            value = value[0]
        if type(value) is not str:
            return None
        parts = value.split(':')
        if len(parts) != 4:
            return None
        try:
            trace_id = long(parts[0], 16)
            span_id = long(parts[1], 16)
            parent_id = long(parts[2], 16)
            flags = int(parts[3], 16)
            if trace_id < 1 or span_id < 1 or parent_id < 0 or flags < 0:
                return None
            if parent_id == 0:
                parent_id = None
            return TraceContext(trace_id=trace_id, span_id=span_id,
                                parent_id=parent_id, flags=flags)
        except ValueError:
            return None



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from concurrent.futures import Future
from opentracing import Tracer


def test_tracer():
    tracer = Tracer()
    span = tracer.start_trace(operation_name='root')
    child = tracer.join_trace(operation_name='child',
                              parent_trace_context=span.trace_context)
    assert span == child
    assert span.trace_context == child.trace_context
    fut = tracer.close()
    assert type(fut) is Future
    fut.result(timeout=1)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from threading import Lock
import opentracing


class TraceContext(opentracing.TraceContext):
    """Base implementation of opentracing.TraceContext.

    This implementation provides thread-safe Trace Attributes propagation
    methods, but does not have any notion of the span identity. It is meant
    for extension by the actual tracing implementations.

    """
    def __init__(self, trace_attributes=None):
        """Initializes this Trace Context with optional Trace Attributes.

        :param trace_attributes: string->string dictionary of Trace
            Attributes. Used as is, without copying, so the caller is
            expected to give ownership of this dictionary.
        :type trace_attributes: dict

        :return: None
        """
        self.trace_attributes = trace_attributes
        self.lock = Lock()

    def normalize_key(self, key):
        """Normalizes the key.

        The keys are always normalized before use by replacing underscores
        with dashes and converting the result to lowercase. For performance
        reasons, the complete validation of the key value to match the
        target rexexp `(?i:[a-z0-9][-a-z0-9]*)` is not performed, it is up
        to the implementation to enforce this.

        :param key: some string key
        :return: normalized key
        """
        return key.replace('_', '-').lower()

    def set_trace_attribute(self, key, value):
        """Implements set_trace_attribute() of opentracing.TraceContext.

        :param key: string key
        :type key: str

        :param value: value of the attribute
        :type value: str

        :rtype : TraceContext
        :return: itself, for chaining the calls.
        """
        assert type(key) is str, 'key must be a string'
        assert type(value) is str, 'value must be a string'

        with self.lock:
            if self.trace_attributes is None:
                self.trace_attributes = dict()
            self.trace_attributes[self.normalize_key(key)] = value
        return self

    def get_trace_attribute(self, key):
        """Implements get_trace_attribute of :type:opentracing.TraceContext.

        :param key: string key
        :return: a Trace Attribute value, or None if no such key was stored.
        """
        with self.lock:
            if self.trace_attributes is None:
                return None
            else:
                return self.trace_attributes.get(self.normalize_key(key))

    def trace_attributes_as_dict(self):
        """Returns a copy of Trace Attributes.

        :return: copy of the Trace Attributes dictionary or None if no
            Trace Attributes was defined in this context.
        """
        with self.lock:
            if self.trace_attributes is None:
                out = None
            else:
                out = dict(self.trace_attributes)
        return out



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
import logging

from concurrent.futures import Future
from . import ioloop_util, thrift

default_logger = logging.getLogger(__name__)


class NullReporter(object):
    """
    Ignores all spans
    """
    def report_span(self, span):
        pass

    def close(self):
        fut = Future()
        fut.set_result(True)
        return fut


class LoggingReporter(NullReporter):
    """
    Logs all spans
    """
    def __init__(self, logger=None):
        self.logger = logger if logger else default_logger

    def report_span(self, span):
        self.logger.warn('Reporting span %s', span)


class CompositeReporter(NullReporter):
    """
    Delegates reporting to one or more underlying reporters.
    """
    def __init__(self, *reporters):
        self.reporters = reporters

    def report_span(self, span):
        for reporter in self.reporters:
            reporter.report_span(span)

    def close(self):
        from threading import Lock
        lock = Lock()
        count = [0]
        future = Future()

        def on_close(_):
            with lock:
                count[0] += 1
                if count[0] == len(self.reporters):
                    future.set_result(True)

        for reporter in self.reporters:
            f = reporter.close()
            f.add_done_callback(on_close)

        return future


class ZipkinReporter(NullReporter):
    """
    Receives completed spans from Tracer and submits them to Collector.
    """

    def __init__(self, channel, buffer_size=10, flush_interval=0,
                 io_loop=None, **kwargs):
        from tornado.ioloop import PeriodicCallback
        from threading import Lock

        self._channel = channel
        self.buffer_size = buffer_size
        self.io_loop = io_loop
        self.logger = kwargs.get('logger', default_logger)

        if self.io_loop is None:
            self.io_loop = ioloop_util.get_io_loop(channel)
        self.io_loop_error_log = False
        self.spans = []
        self.spans_lock = Lock()
        if self.io_loop and flush_interval > 0:
            self.periodic = PeriodicCallback(callback=self.flush,
                                             callback_time=flush_interval*1000,  # to milliseconds
                                             io_loop=io_loop)
            self.periodic.start()
        else:
            self.periodic = None

    def report_span(self, span):
        with self.spans_lock:
            self.spans.append(span)
            size = len(self.spans)
        if size >= self.buffer_size:
            self.flush()

    def close(self):
        if self.periodic:
            self.periodic.stop()
        return self.flush()

    def flush(self):
        if self.io_loop is None:
            if not self.io_loop_error_log:
                self.logger.error('Reporter has no IOLoop')
                self.io_loop_error_log = True
                return ioloop_util.future_result(False)
        return ioloop_util.submit(self._flush, io_loop=self.io_loop)

    def _flush(self):

        def submit_callback(future):
            ex = future.exception()
            if ex:
                self.logger.error(
                    'Failed to submit trace to Collector: %s', ex)

        with self.spans_lock:
            spans = self.spans
            self.spans = []
        if len(spans) > 0:
            try:
                request = thrift.make_submit_batch_request(spans)
                fut = self._channel.thrift(request)
                fut.add_done_callback(submit_callback)
                return fut
            except Exception as e:
                self.logger.error('Failed to submit trace to tcollector: %s', e)
                return ioloop_util.future_exception(e)
        return ioloop_util.future_result(True)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
import threading

import opentracing
from opentracing.ext import tags
from .glossary import SERVER_SEND, SERVER_RECV, CLIENT_SEND, CLIENT_RECV
from .glossary import SERVER_ADDR, CLIENT_ADDR
from . import thrift

SAMPLED_FLAG = 0x01
DEBUG_FLAG = 0x02


class Span(opentracing.Span):
    """Implements opentracing.Span with Zipkin semantics. """

    def __init__(self, trace_context, tracer, operation_name, is_client=False):
        super(Span, self).__init__(trace_context)
        self.tracer = tracer
        self.operation_name = operation_name
        self.is_client = is_client
        self.peer = None
        self.ts = None
        self.duration = None
        self.update_lock = threading.Lock()
        # we store tags and logs as Zipkin native thrift BinaryAnnotation and
        # Annotation structures, to avoid creating intermediate objects
        self.tags = []
        self.logs = []

    def start_child(self, operation_name, tags=None):
        """Implements start_child() of opentracing.Span"""
        child_ctx, child_tags = self.tracer.trace_context_source. \
            new_child_trace_context(parent_trace_context=self.trace_context)

        span = self.tracer.create_span(operation_name=operation_name,
                                       trace_context=child_ctx,
                                       is_client=True,
                                       tags=child_tags)
        if tags:
            span.add_tags(tags)
        return span

    def finish(self):
        """Implements finish() of opentracing.Span"""
        if self.is_sampled():
            ts = self.tracer.timestamp()
            self.duration = ts - self.ts
            end_event = CLIENT_RECV if self.is_client else SERVER_SEND
            end_event = thrift.make_event(timestamp=ts, name=end_event)

            start_event = CLIENT_SEND if self.is_client else SERVER_RECV
            start_event = thrift.make_event(timestamp=self.ts,
                                            name=start_event)

            if self.peer:
                host = thrift.make_endpoint(
                    ipv4=self.peer.get('ipv4', 0),
                    port=self.peer.get('port', 0),
                    service_name=self.peer.get('service_name', ''))
                addr = SERVER_ADDR if self.is_client else CLIENT_ADDR
                peer = thrift.make_peer_address_tag(key=addr, host=host)
            else:
                peer = None

            with self.update_lock:
                self.logs.append(start_event)
                self.logs.append(end_event)
                if peer:
                    self.tags.append(peer)

            self.tracer.report_span(self)

    def add_tag(self, key, value):
        """Implements add_tag() method of opentracing.Span"""
        if self.is_sampled():
            special = SPECIAL_TAGS.get(key, None)
            if special is not None and callable(special):
                special(self, value)
            else:
                tag = thrift.make_string_tag(key, value)
                with self.update_lock:
                    self.tags.append(tag)
        return self

    def add_tags(self, tags):
        """Implements add_tags() method of opentracing.Span"""
        if self.is_sampled():
            for k, v in tags.itemiter():
                self.add_tag(k, v)
        return self

    def info(self, message, *payload):
        self.add_log(message, False, *payload)

    def error(self, message, *payload):
        self.add_log(message, True, *payload)

    def add_log(self, message, is_error, *payload):
        """Internal method"""
        if self.is_sampled():
            timestamp = self.tracer.timestamp()
            log = thrift.make_event(timestamp, message)
            # Since Zipkin format does not support logs with payloads,
            # we convert the payload to a tag with the same name as log.
            # Also, if the log was an error, we add yet another tag.
            if payload:
                tag = thrift.make_string_tag(message, str(payload))
            else:
                tag = None
            if is_error:
                err = thrift.make_string_tag(message, 'error')
            else:
                err = None
            with self.update_lock:
                self.logs.append(log)
                if tag:
                    self.tags.append(tag)
                if err:
                    self.tags.append(err)

    def is_sampled(self):
        return self.trace_context.flags & SAMPLED_FLAG == SAMPLED_FLAG

    def is_debug(self):
        return self.trace_context.flags & DEBUG_FLAG == DEBUG_FLAG

    def __str__(self):
        from .marshaling import TraceContextEncoder

        id_str = TraceContextEncoder.id_to_string(self.trace_context)
        return "%s.%s %s" % (self.tracer.service_name,
                             self.operation_name,
                             id_str)


def _peer_service(span, value):
    with span.update_lock:
        if span.peer is None:
            span.peer = {'service_name': value}
        else:
            span.peer['service_name'] = value


def _peer_host_ipv4(span, value):
    with span.update_lock:
        if span.peer is None:
            span.peer = {'ipv4': value}
        else:
            span.peer['ipv4'] = value


def _peer_port(span, value):
    with span.update_lock:
        if span.peer is None:
            span.peer = {'port': value}
        else:
            span.peer['port'] = value


SPECIAL_TAGS = {
    tags.PEER_SERVICE: _peer_service,
    tags.PEER_HOST_IPV4: _peer_host_ipv4,
    tags.PEER_PORT: _peer_port,
}



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from __future__ import absolute_import

# Here we define standard names for tags that can be added to spans by the
# instrumentation code. The actual tracing systems are not required to
# retain these as tags in the stored spans if they have other means of
# representing the same data. For example, the SPAN_KIND='s' (server) can be
# inferred from a Zipkin span by the presence of ss/sr annotations.

# PEER_* tags can be emitted by either client-side of server-side to describe
# the other side/service in a peer-to-peer communications, like an RPC call.

# PEER_SERVICE records the service name of the peer
PEER_SERVICE = 'peer.service'

# PEER_HOSTNAME records the host name of the peer
PEER_HOSTNAME = 'peer.hostname'

# PEER_HOST_IPV4 records IP v4 host address of the peer
PEER_HOST_IPV4 = 'peer.ipv4'

# PEER_HOST_IPV6 records IP v6 host address of the peer
PEER_HOST_IPV6 = 'peer.ipv6'

# PEER_PORT records port number of the peer
PEER_PORT = 'peer.port'

# SPAN_KIND hints at relationship between spans, e.g. client/server
SPAN_KIND = 'span.kind'

# Marks a span representing the client-side of an RPC or other remote call
SPAN_KIND_RPC_CLIENT = 'c'

# Marks a span representing the server-side of an RPC or other remote call
SPAN_KIND_RPC_SERVER = 's'



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
import logging
import time

MAX_ID_BITS = 63

default_logger = logging.getLogger(__name__)


class Sampler(object):
    """
    Sampler is responsible for deciding if a particular span should be
    "sampled", i.e. recorded in permanent storage.
    """

    def is_sampled(self, trace_id):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()


class ConstSampler(Sampler):
    """
    ConstSampler always returns the same decision.
    """

    def __init__(self, decision):
        self.decision = decision

    def is_sampled(self, trace_id):
        return self.decision

    def close(self):
        pass

    def __str__(self):
        return 'ConstSampler(%s)' % self.decision


class ProbabilisticSampler(Sampler):
    """
    A sampler that randomly samples a certain percentage of traces specified
    by the samplingRate, in the range between 0.0 and 1.0.

    It relies on the fact that new trace IDs are 64bit random numbers
    themselves, thus making the sampling decision without generating a new
    random number, but simply calculating if traceID < (samplingRate * 2^64).
    Note that we actually ignore (zero out) the most significant bit.
    """

    def __init__(self, rate):
        assert 0.0 <= rate <= 1.0, "Sampling rate must be between 0.0 and 1.0"
        self.rate = rate
        self.max_number = 1 << MAX_ID_BITS
        self.boundary = rate * self.max_number

    def is_sampled(self, trace_id):
        return trace_id < self.boundary

    def close(self):
        pass

    def __str__(self):
        return 'ProbabilisticSampler(%s)' % self.rate


class RateLimitingSampler(Sampler):
    """
    Samples at most max_traces_per_second. The distribution of sampled
    traces follows burstiness of the service, i.e. a service with uniformly
    distributed requests will have those requests sampled uniformly as well,
    but if requests are bursty, especially sub-second, then a number of
    sequential requests can be sampled each second.
    """

    def __init__(self, max_traces_per_second=10):
        assert max_traces_per_second >= 0, \
            'max_traces_per_second must not be negative'
        self.credits_per_second = max_traces_per_second
        self.balance = max_traces_per_second
        self.last_tick = self.timestamp()
        self.item_cost = 1

    def is_sampled(self, trace_id):
        current_time = self.timestamp()
        elapsed_time = current_time - self.last_tick
        self.last_tick = current_time
        self.balance += elapsed_time * self.credits_per_second
        if self.balance > self.credits_per_second:
            self.balance = self.credits_per_second
        if self.balance >= self.item_cost:
            self.balance -= self.item_cost
            return True
        return False

    def close(self):
        pass

    @staticmethod
    def timestamp():
        return time.time()

    def __str__(self):
        return 'RateLimitingSampler(%s)' % self.credits_per_second



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import


class RawSpan(object):
    """
    Represents all accumulated state associated with a finished Span.
    """
    def __init__(self, trace_context, operation_name):
        self.trace_context = trace_context
        self.operation_name = operation_name
        self.start_time = None
        self.duration = None
        self.tags = None  # array of (key,vale) pairs when initialized
        self.logs = None  # array of RawLog's when initialized

    def add_tag(self, key, value):
        if self.tags is None:
            self.tags = []
        self.tags.append((key, value))

    def add_log(self, log):
        if self.logs is None:
            self.logs = []
        self.logs.append(log)


class RawLog(object):
    """
    Encapsulates a log attached to a Span.
    """
    def __init__(self, timestamp, is_error, message, *args):
        self.timestamp = timestamp
        self.is_error = is_error
        self.message = message
        self.payload = args



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import


class TraceContext(object):
    """Represents the identity of a given Span and Trace Attributes.

    The span identity is embedded in this context in implementation-specific
    format. The contents are serialized on the wire during inter-process calls.

    Trace Attributes can be added to the context and will be available to all
    future children of this span, but will not propagate up the call tree.

    The more attributes are added to the context, the heavier it makes the
    requests on the wire, so it must be used with caution.
    """

    def set_trace_attribute(self, key, value):
        """Stores Trace Attribute in the context as a key/value pair.

        Enables powerful distributed context propagation functionality where
        arbitrary application data can be carried along the full path of
        request execution throughout the system.

        Note 1: attributes are only propagated to the future children of the
        current trace context.

        Note 2: attributes are sent in-band with every subsequent local and
        remote calls, so this feature must be used with care.

        Note 3: keys are case-insensitive, to allow propagation  via HTTP
        headers. Keys must match rexep `(?i:[a-z0-9][-a-z0-9]*)`

        :param key: trace attribute key
        :type key: str

        :param value: trace attribute value
        :type value: str

        :rtype : TraceContext
        :return: itself, for chaining the calls.
        """
        return self

    def get_trace_attribute(self, key):
        """Retrieves value of the Trace Attribute with the given key.

        Key is case-insensitive.

        :param key: key of the Trace Attribute
        :type key: str

        :rtype : str
        :return: value of the Trace Attribute with given key, or None.
        """
        return None


class TraceContextEncoder(object):
    """Encodes a trace context in binary or text formats.

    The encoder is expected to serialize trace contexts into a pair of
    values representing separately the trace context / span identity,
    and the trace attributes. This is done specifically for binary protocols
    that may represent tracing identity in a dedicated fixed-length slot
    in the binary message format, so that it can be inspected efficiently
    by the middleware / routing layers without parsing the whole message.
    """

    def trace_context_to_binary(self, trace_context):
        """Converts trace context to a pair of bytearray's representing
        separately span identity and trace attributes.

        :param trace_context: trace context to encode
        :type trace_context: TraceContext

        :rtype (bytearray, bytearray or None)
        :return: a pair of bytearray's, first representing the span identity,
            the second representing Trace Attributes (can be None).
        """
        return bytearray(), None

    def trace_context_to_text(self, trace_context):
        """Converts trace context to a pair of dictionaries representing
        separately span identity and Trace Attributes.

        :param trace_context: trace context to encode
        :type trace_context: TraceContext

        :rtype (dict, dict or None)
        :return: a pair of string->string dictionaries, first representing the
            span identity, the second representing attributes (can be None).
        """
        return dict(), None


class TraceContextDecoder(object):
    """Decodes a trace context from binary or text formats."""

    def trace_context_from_binary(self, trace_context_id, trace_attributes):
        """Converts encoded binary data into a TraceContext.

        Singe this is a reference no-op implementation, it always returns
        two same singleton context.

        :param trace_context_id: span identity encoded as bytearray
        :param trace_attributes: attributes encoded as bytearray

        :return: a reconstructed TraceContext.
        """
        return TraceContextSource.singleton_noop_trace_context

    def trace_context_from_text(self, trace_context_id, trace_attributes):
        """Converts encoded string data into a TraceContext.

        Singe this is a reference no-op implementation, it always returns
        two same singleton context.

        :param trace_context_id: span identity encoded as string->string map
        :param trace_attributes: attributes encoded as string->string map

        :return: a reconstructed TraceContext.
        """
        return TraceContextSource.singleton_noop_trace_context


class TraceContextSource(TraceContextEncoder, TraceContextDecoder):
    """Knows how to create new TraceContext

    Since this is a reference no-op implementation, it always returns the
    same singleton context `singleton_noop_trace_context`.
    """

    singleton_noop_trace_context = TraceContext()

    def new_root_trace_context(self):
        """Creates a new TraceContext without parent and with unique identity.

        As the context has no parent, it implicitly starts a new trace.

        This is a no-op implementation that always returns the same context.

        :return: a new trace context without parent.
        """
        return TraceContextSource.singleton_noop_trace_context

    def new_child_trace_context(self, parent_trace_context):
        """Creates a child context from given parent context.

        Since this is a no-op implementation, it returns itself as a child.
        Actual implementation must create a new context that is a logical
        child of the current context.

        :param parent_trace_context: parent context

        :rtype : (TraceContext, dict or None)
        :return: a pair of (child_context, child_tags), where child_tags is
            an optional dictionary of tags to be added to the child span.
        """
        return TraceContextSource.singleton_noop_trace_context, None



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import socket
import struct

from .zipkin_thrift import zipkin_collector


def ipv4_to_int(ipv4):
    if ipv4 == 'localhost':
        ipv4 = '127.0.0.1'
    elif ipv4 == '::1':
        ipv4 = '127.0.0.1'
    try:
        return struct.unpack('!i', socket.inet_aton(ipv4))[0]
    except:
        return 0


def port_to_int(port):
    if type(port) is str:
        if port.isdigit():
            port = int(port)
    if type(port) is int:
        # zipkinCore.thrift defines port as i16, which is signed,
        # therefore we convert ephemeral ports as negative ints
        if port >= 32768:
            port -= (1 << 16)
        return port
    return None


def make_endpoint(ipv4, port, service_name):
    if type(ipv4) is str:
        ipv4 = ipv4_to_int(ipv4)
    port = port_to_int(port)
    if port is None:
        port = 0
    return zipkin_collector.Endpoint(ipv4, port, service_name.lower())


def make_string_tag(key, value):
    return zipkin_collector.BinaryAnnotation(
        key, value, zipkin_collector.AnnotationType.STRING)


def make_peer_address_tag(key, host):
    """
    Used for Zipkin binary annotations like CA/SA (client/server address).
    They are modeled as Boolean type with '0x01' as the value.
    """
    return zipkin_collector.BinaryAnnotation(
        key, '0x01', zipkin_collector.AnnotationType.BOOL, host)


def make_event(timestamp=None, name=None):
    return zipkin_collector.Annotation(
        timestamp=long(timestamp), value=name, host=None)


def make_zipkin_spans(spans):
    zipkin_spans = []
    for span in spans:
        ctx = span.trace_context
        endpoint = make_endpoint(ipv4=span.tracer.ip_address,
                                 port=0,  # span.port,
                                 service_name=span.tracer.service_name)
        # TODO extend Zipkin Thrift and pass endpoint once only
        for event in span.logs:
            event.host = endpoint
        zipkin_span = zipkin_collector.Span(
            trace_id=ctx.trace_id,
            name=span.operation_name,
            id=ctx.span_id,
            parent_id=ctx.parent_id,
            annotations=span.logs,
            binary_annotations=span.tags,
            debug=span.is_debug(),
            timestamp=long(span.ts),
            duration=long(span.duration)
        )
        zipkin_spans.append(zipkin_span)
    return zipkin_spans


def make_submit_batch_request(spans):
    zipkin_spans = make_zipkin_spans(spans)
    return zipkin_collector.ZipkinCollector.submitZipkinBatch(zipkin_spans)



# Copyright (c) 2016 The OpenTracing Authors.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from __future__ import absolute_import


class APICompatibilityCheckMixin(object):
    """
    A mixin class for validation that a given tracer implementation
    satisfies the requirements of the OpenTracing API.
    """
    def tracer(self):
        raise NotImplementedError('Subclass must implement tracer()')

    def check_trace_attribute_values(self):
        """If true, the test will validate trace attributes by
        storing and retrieving them from the trace context. If false,
        it will only attempt to store and retrieve the attributes to check
        the API compliance, but not actually validate stored values.
        The latter mode is only useful for no-op tracer.
        """
        return True

    def test_start_trace(self):
        tracer = self.tracer()
        span = tracer.start_trace(operation_name='Fry')
        span.finish()
        with tracer.start_trace(operation_name='Fry',
                                tags={'birthday': 'August 14 1974'}) as span:
            span.info('birthplace',
                      {'hospital': 'Brooklyn Pre-Med Hospital',
                       'city': 'Old New York'})
        tracer.close()

    def test_join_trace(self):
        tracer = self.tracer()
        trace_context = tracer.new_root_trace_context()
        assert trace_context is not None
        span = tracer.join_trace(operation_name='Leela',
                                 parent_trace_context=trace_context)
        span.finish()
        span = tracer.join_trace(operation_name='Leela',
                                 parent_trace_context=trace_context,
                                 tags={'birthplace': 'sewers'})
        span.finish()
        tracer.close()

    def test_child_span(self):
        parent = self.tracer().start_trace(operation_name='Farnsworth')
        with parent as p:
            child = p.start_child(operation_name='Cubert')
            child.finish()

    def test_span_as_context_manager(self):
        finish = {'called': False}

        def mock_finish(*_):
            finish['called'] = True

        with self.tracer().start_trace(operation_name='antiquing') as span:
            setattr(span, 'finish', mock_finish)
        assert finish['called'] is True

        # now try with exception
        finish['called'] = False
        try:
            with self.tracer().start_trace(operation_name='antiquing') as span:
                setattr(span, 'finish', mock_finish)
                raise ValueError()
        except ValueError:
            assert finish['called'] is True
        else:
            raise AssertionError('Expected ValueError')  # pragma: no cover

    def test_span_tags(self):
        span = self.tracer().start_trace(operation_name='Farnsworth')
        span.add_tag('birthday', '9 April, 2841')
        span.add_tag('loves', 'different lengths of wires')
        span.finish()

    def test_span_logs(self):
        span = self.tracer().start_trace(operation_name='Fry')
        span.info('frozen', {'year': 1999, 'place': 'Cryogenics Labs'})
        span.error('defrosted', {'year': 2999})

    def test_trace_attributes(self):
        trace_context = self.tracer().new_root_trace_context()
        trace_context.set_trace_attribute('Kiff-loves', 'Amy')
        val = trace_context.get_trace_attribute('kiff-Loves')  # case change
        if self.check_trace_attribute_values():
            assert 'Amy' == val

    def test_trace_context_source_child(self):
        trace_context = self.tracer().new_root_trace_context()
        child_context, tags = self.tracer().new_child_trace_context(
            parent_trace_context=trace_context
        )
        child_context.set_trace_attribute('child', 'Igner')
        assert tags is None or type(tags) is dict

    def test_text_codecs(self):
        span = self.tracer().start_trace(operation_name='Bender')
        trace_context = span.trace_context
        dict_trace_id, dict_attrs = self.tracer().trace_context_to_text(
            trace_context=trace_context
        )
        assert type(dict_trace_id) is dict
        assert dict_attrs is None or type(dict_attrs) is dict
        trace_context = self.tracer().trace_context_from_text(
            trace_context_id=dict_trace_id,
            trace_attributes=dict_attrs
        )
        trace_context.set_trace_attribute('middle-name', 'Rodriguez')

    def test_binary_codecs(self):
        span = self.tracer().start_trace(operation_name='Bender')
        trace_context = span.trace_context
        bin_trace_id, bin_attrs = self.tracer().trace_context_to_binary(
            trace_context=trace_context
        )
        assert type(bin_trace_id) is bytearray
        assert bin_attrs is None or type(bin_attrs) is bytearray
        trace_context = self.tracer().trace_context_from_binary(
            trace_context_id=bin_trace_id,
            trace_attributes=bin_attrs
        )
        trace_context.set_trace_attribute('middle-name', 'Rodriguez')



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from concurrent.futures import Future
from .span import Span
from .context import TraceContextSource
from .context import TraceContextEncoder, TraceContextDecoder


class Tracer(TraceContextSource, object):
    """Tracer is the entry point API between instrumentation code and the
    tracing implementation.

    This implementation both defines the public Tracer API, and provides
    a default no-op behavior.
    """

    singleton_noop_span = Span(TraceContextSource.singleton_noop_trace_context)

    def start_trace(self, operation_name, tags=None):
        """Starts a new trace and creates a new root span.

        This method should be used by services that are instrumented for
        tracing but did not receive trace ID from upstream request.

        :param operation_name: the service's own name for the end-point
            that received the request represented by this trace and span.
            The domain of names must be limited, e.g. do not use UUIDs or
            entity IDs or timestamps as part of the name.
        :param tags: optional dictionary of Span Tags. The caller is
            expected to give up ownership of that dictionary, because the
            Tracer may use it as is to avoid extra data copying.

        :return: a new root Span
        """
        return Tracer.singleton_noop_span

    def join_trace(self, operation_name, parent_trace_context, tags=None):
        """Joins a trace started elsewhere and creates a new span as a
        child of the given parent trace context.

        This method should be used by services that receive tracing info
        from upstream.

        :param operation_name: the service's own name for the end-point
            that received the request represented by this trace and span.
            The domain of names must be limited, e.g. do not use UUIDs or
            entity IDs or timestamps as part of the name.
        :param parent_trace_context: Trace Context of a client span started
            elsewhere and whose trace we're joining.
        :param tags: optional dictionary of Span Tags. The caller is
            expected to give up ownership of that dictionary, because the
            Tracer may use it as is to avoid extra data copying.

        :return: a new Span
        """
        return Tracer.singleton_noop_span

    def close(self):
        """Performs a clean shutdown of the tracer, flushing any traces that
        may have been buffered in memory.

        :return: Returns a :py:class:futures.Future
        """
        fut = Future()
        fut.set_result(True)
        return fut



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
import threading
import random
import time

import os
import opentracing.context
import opentracing.standard.context

from .span import SAMPLED_FLAG, DEBUG_FLAG
from .sampler import MAX_ID_BITS


class TraceContext(opentracing.standard.context.TraceContext):
    """
    Zipkin-style subclass of standard TraceContext
    """

    def __init__(self, trace_id, span_id, parent_id, flags,
                 trace_attributes=None):
        """
        Constructor of a Trace Context.
        :param trace_id:
        :param span_id:
        :param parent_id:
        :param flags:
        :param trace_attributes: string->string dictionary of Trace Attributes.
            Used as is, without copying, so the caller is expected to give
            ownership of this dictionary.
        :return: new TraceContext
        """
        super(TraceContext, self).__init__(trace_attributes=trace_attributes)
        self.trace_id = trace_id
        self.span_id = span_id
        self.parent_id = parent_id
        self.flags = flags

    def span_id_as_tuple(self):
        """
        Method used in testing
        :return: returns a tuple (trace_id, span_id, parent_id, flags)
        """
        return self.trace_id, self.span_id, self.parent_id, self.flags

    def __str__(self):
        from .encoding import TraceContextEncoder

        return TraceContextEncoder.id_to_string(self)

    def append_trace_attributes(self, trace_attributes):
        """Adds extra Trace Attributes to the current context

        :param: extra Trace Attributes dictionary
        :return: None
        """
        with self.lock:
            if self.trace_attributes is None:
                self.trace_attributes = {}
            for key in trace_attributes.keys():
                norm_key = self.normalize_key(key)
                self.trace_attributes[norm_key] = trace_attributes[key]


class TraceContextSource(opentracing.context.TraceContextSource):
    """
    Zipkin-style TraceContextSource.
    """

    def __init__(self, sampler):
        self.sampler = sampler
        self.random = random.Random(time.time() * (os.getpid() or 1))
        self.sampler_lock = threading.Lock()

    def random_id(self):
        return self.random.getrandbits(MAX_ID_BITS)

    def new_root_trace_context(self, debug=False):
        with self.sampler_lock:
            trace_id = self.random_id()
            if debug:
                flags = SAMPLED_FLAG + DEBUG_FLAG
            elif self.sampler.is_sampled(trace_id):
                flags = SAMPLED_FLAG
            else:
                flags = 0
        return TraceContext(trace_id=trace_id, span_id=trace_id,
                            parent_id=None, flags=flags)

    def new_child_trace_context(self, parent_trace_context):
        trace_attributes = parent_trace_context.trace_attributes_as_dict()
        ctx = TraceContext(trace_id=parent_trace_context.trace_id,
                           span_id=self.random_id(),
                           parent_id=parent_trace_context.span_id,
                           flags=parent_trace_context.flags,
                           trace_attributes=trace_attributes)
        return ctx, None

    def close(self):
        self.sampler.close()



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
import mock
from opentracing import Span, TraceContextSource
from opentracing.ext import tags


def test_span():
    ctx = TraceContextSource().new_root_trace_context()
    span = Span(trace_context=ctx)
    assert span.trace_context == ctx
    child = span.start_child(operation_name='test')
    assert span == child
    child.info('cache hit', 'arg1', 'arg2')
    child.error('cache miss', 'arg1', 'arg2')

    with mock.patch.object(span, 'finish') as finish:
        with mock.patch.object(span, 'error') as error:
            try:
                with span:
                    raise ValueError()
            except ValueError:
                pass
            assert finish.call_count == 1
            assert error.call_count == 1

    with mock.patch.object(span, 'finish') as finish:
        with mock.patch.object(span, 'error') as error:
            with span:
                pass
            assert finish.call_count == 1
            assert error.call_count == 0

    span.add_tag('x', 'y').add_tag('z', 1)
    span.add_tag(tags.PEER_SERVICE, 'test-service')
    span.add_tag(tags.PEER_HOST_IPV4, 127 << 24 + 1)
    span.add_tag(tags.PEER_HOST_IPV6, '::')
    span.add_tag(tags.PEER_HOSTNAME, 'uber.com')
    span.add_tag(tags.PEER_PORT, 123)
    span.finish()



from setuptools import setup


setup(
    name='opentracing',
    version='0.6.1',
    author='OpenTracing Authors',
    author_email='shkuro@gmail.com',
    description='OpenTracing API for Python. See documentation at http://opentracing.io',
    license='MIT',
    url='https://github.com/opentracing/api-python',
    download_url = 'https://github.com/opentracing/api-python/tarball/0.5.0',
    keywords=['opentracing'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=['opentracing'],
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=[
        'futures',
    ],
    extras_require={
        'tests': [
            'doubles',
            'flake8',
            'flake8-quotes',
            'mock<1.1.0',
            'pytest',
            'pytest-cov',
            'pytest-mock',
            'Sphinx',
            'sphinx_rtd_theme',
        ]
    },
)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from opentracing import TraceContext, TraceContextSource
from opentracing import TraceContextEncoder, TraceContextDecoder


def test_context():
    ctx = TraceContext()
    assert ctx.get_trace_attribute('x') is None
    ctx.set_trace_attribute('X_y', 'value').\
        set_trace_attribute('ZZ', 'value2')
    assert ctx.get_trace_attribute('X_y') is None


def test_context_source():
    singleton = TraceContextSource.singleton_noop_trace_context
    source = TraceContextSource()
    assert source.new_root_trace_context() == singleton
    child, attr = source.new_child_trace_context(
        parent_trace_context=singleton)
    assert child == singleton
    assert attr is None


def test_encoder():
    singleton = TraceContextSource.singleton_noop_trace_context
    e = TraceContextEncoder()
    x, y = e.trace_context_to_binary(trace_context=singleton)
    assert x == bytearray()
    assert y is None
    x, y = e.trace_context_to_text(trace_context=singleton)
    assert x == {}
    assert y is None


def test_decoder():
    singleton = TraceContextSource.singleton_noop_trace_context
    d = TraceContextDecoder()
    ctx = d.trace_context_from_binary(trace_context_id=None,
                                      trace_attributes=None)
    assert singleton == ctx
    ctx = d.trace_context_from_text(trace_context_id=None,
                                    trace_attributes=None)
    assert singleton == ctx



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
import time

from tchannel.net import local_ip

import opentracing
from . import glossary
from .glossary import TRACE_ATTRIBUTES_HEADER_PREFIX
from .marshaling import TraceContextEncoder, TraceContextDecoder
from .span import Span
from .version import __version__
from .thrift import ipv4_to_int


class Tracer(opentracing.Tracer):
    def __init__(self, service_name, reporter, trace_context_source):
        self.service_name = service_name
        self.reporter = reporter
        self.trace_context_source = trace_context_source
        self.ip_address = ipv4_to_int(local_ip())
        self.encoder = TraceContextEncoder(
            trace_id_header=glossary.TRACE_ID_HEADER,
            trace_attributes_header_prefix=TRACE_ATTRIBUTES_HEADER_PREFIX)
        self.decoder = TraceContextDecoder(
            trace_id_header=glossary.TRACE_ID_HEADER,
            trace_attributes_header_prefix=TRACE_ATTRIBUTES_HEADER_PREFIX)

    def start_trace(self, operation_name, tags=None):
        """Implements start_trace of opentracing.Tracer."""
        trace_context = self.trace_context_source.new_root_trace_context(debug)
        return self.create_span(operation_name=operation_name,
                                trace_context=trace_context,
                                is_client=False, tags=tags)

    def join_trace(self, operation_name, parent_trace_context, tags=None):
        """Implements join_trace of opentracing.Tracer"""
        # NOTE: Zipkin-specific behavior - server joins the same span that
        # was started by the client.
        return self.create_span(operation_name=operation_name,
                                trace_context=parent_trace_context,
                                is_client=False, tags=tags)

    def close(self):
        """Performs a clean shutdown of the tracer, flushing any traces that
        may be buffered in memory.

        :return: Returns a concurrent.futures.Future that indicates if the
            flush has been completed.
        """
        self.trace_context_source.close()
        return self.reporter.close()

    def create_span(self, operation_name, trace_context, is_client, tags=None):
        """
        Internal method
        """
        span = Span(trace_context=trace_context,
                    tracer=self,
                    operation_name=operation_name,
                    is_client=is_client)
        if tags:
            for k, v in tags.iteritems():
                span.add_tag(k, v)
        span.ts = self.timestamp()
        span.add_tag(key='tracing.client', value='Python-%s' % __version__)
        return span

    def report_span(self, span):
        self.reporter.report_span(span)

    @staticmethod
    def timestamp():
        """Returns current time in microseconds."""
        return time.time() * 1000000

    def trace_context_to_binary(self, trace_context):
        return self.encoder.trace_context_to_binary(
            trace_context=trace_context
        )

    def trace_context_to_text(self, trace_context):
        return self.encoder.trace_context_to_text(
            trace_context=trace_context
        )

    def trace_context_from_binary(self, trace_context_id,
                                  trace_attributes):
        return self.decoder.trace_context_from_binary(
            trace_context_id=trace_context_id,
            trace_attributes=trace_attributes
        )

    def trace_context_from_text(self, trace_context_id,
                                trace_attributes):
        return self.decoder.trace_context_from_text(
            trace_context_id=trace_context_id,
            trace_attributes=trace_attributes
        )
